from typing import Optional
import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from torchok.constructor import HEADS
from torchok.models.base import BaseModel


@HEADS.register_class
class ArcFaceHead(BaseModel):
    """Implement of arc margin distance. Classification head.

    ArcFace paper: https://arxiv.org/pdf/1801.07698.pdf
    Code: https://github.com/ronghuaiyang/arcface-pytorch/blob/master/models/metrics.py
    """

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 scale: float = None,
                 margin: float = None,
                 easy_margin: bool = False,
                 dynamic_margin: bool = False,
                 num_warmup_steps: int = None,
                 min_margin: float = None):
        """Init ArcFaceHead class.

        Args:
            in_channels: Size of each input sample.
            num_classes: number of classes.
            scale: Feature scale.
            margin: Angular margin.
            easy_margin: Easy margin.
            dynamic_margin: If True margin will increase from min_margin to margin,
                step by step(num_warmup_steps times).
            num_warmup_steps: Number steps with dynamic margin.
            min_margin: Initial margin in dynamic_margin mode.

        Raises:
            ValueError: if num_warmup_steps or min_margin is None, when `dynamic_margin` is True.
        """
        super().__init__(in_channels, out_channels=num_classes)

        if scale is None:
            p = .999
            c_1 = (num_classes - 1)
            scale = c_1 / num_classes * math.log(c_1 * p / (1 - p)) + 1

        if margin is None:
            if in_channels == 2:
                margin = .9 - math.cos(2 * math.pi / num_classes)
            else:
                margin = .5 * num_classes / (num_classes - 1)

        self.__dynamic_margin = dynamic_margin

        if self.__dynamic_margin:
            if num_warmup_steps is None or not isinstance(num_warmup_steps, int):
                raise ValueError('`num_warmup_steps` must be positive int when `dynamic_margin` is True')
            if min_margin is None:
                raise ValueError('`min_margin` must be float when `dynamic_margin` is True')
            self.__num_warmup_steps = num_warmup_steps
            self.__min_margin = min_margin
            self.__max_margin = margin
            self.__margin = min_margin
            self.__register_buffer('step', torch.tensor(0))
        else:
            self.__margin = margin

        self.__scale = scale
        self.__easy_margin = easy_margin
        self.__cos_m = torch.scalar_tensor(math.cos(self.__margin))
        self.__sin_m = torch.scalar_tensor(math.sin(self.__margin))
        self.__th = torch.scalar_tensor(math.cos(math.pi - self.__margin))
        self.__mm = torch.scalar_tensor(math.sin(math.pi - self.__margin) * self.__margin)

        self.__weight = nn.Parameter(torch.zeros(num_classes, in_channels), requires_grad=True)

        nn.init.xavier_uniform_(self.__weight)

    def __update_margin(self) -> None:
        if self.__dynamic_margin and self.__step <= self.__num_warmup_steps:
            frac = self.__step / self.__num_warmup_steps
            margin_gap = self.__max_margin - self.__min_margin
            self.__margin = self.__min_margin + frac * margin_gap
            self.__cos_m = torch.scalar_tensor(math.cos(self.__margin))
            self.__sin_m = torch.scalar_tensor(math.sin(self.__margin))
            self.__th = torch.scalar_tensor(math.cos(math.pi - self.__margin))
            self.__mm = torch.scalar_tensor(math.sin(math.pi - self.__margin) * self.__margin)
            self.__step += 1

    def __add_margin(self, cosine: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        sine = torch.sqrt((1.0 - torch.pow(cosine, 2)).clamp(0, 1))
        phi = (cosine * self.__cos_m - sine * self.__sin_m).type(cosine.dtype)
        if self.__easy_margin:
            phi = torch.where(cosine > 0, phi, cosine)
        else:
            phi = torch.where(cosine > self.__th, phi, cosine - self.__mm)

        one_hot = torch.zeros_like(cosine)
        one_hot.scatter_(1, target.view(-1, 1).long(), 1)
        output = torch.where(one_hot == 1, phi, cosine)
        output *= self.__scale

        return output

    def forward(self, input: torch.Tensor, target: Optional[torch.Tensor] = None) -> torch.Tensor:
        """Forward method.

        Args:
            input: Input tensor.
            target: Target tensor. May be None, if training mode off (inference stage).

        Raises:
            ValueError: If training mode on and target is None.
        """
        if not self.training:
            return F.linear(input, self.__weight)
        elif target is None:
            raise ValueError('Target is None in training mode.')

        x = F.normalize(input)
        weight = F.normalize(self.__weight)
        cosine = F.linear(x, weight)
        output = self.__add_margin(cosine, target)
        self.__update_margin()

        return output

    @property
    def margin(self) -> float:
        """Angular margin."""
        return self.__margin

    @property
    def scale(self) -> float:
        """Feature scale."""
        return self.__scale

    @property
    def dynamic_margin(self) -> bool:
        """Dynamic margin mode."""
        return self.__dynamic_margin

    @property
    def num_warmup_steps(self) -> Optional[int]:
        """It's number of warm-up steps."""
        return self.__num_warmup_steps

    @property
    def min_margin(self) -> Optional[float]:
        """It's initial margin in `dynamic_margin` mode."""
        return self.__min_margin

    @property
    def weights(self) -> torch.Tensor:
        """It's ArcFaceHead weights."""
        return self.__weight



from pathlib import Path
from typing import Union, Optional
import pickle
import numpy as np


import torch
from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose
from torchok.data.datasets.base import ImageDataset
from torchvision.datasets.utils import check_integrity, download_and_extract_archive

from torchok.constructor import DATASETS


@DATASETS.register_class
class CIFAR10(ImageDataset):
    """A class represent cifar10 dataset."""
    base_folder = 'cifar-10-batches-py'
    url = 'https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz'
    filename = 'cifar-10-python.tar.gz'
    tgz_md5 = 'c58f30108f718f92721af3b95e74349a'
    train_list = [
        ['data_batch_1', 'c99cafc152244af753f735de768cd75f'],
        ['data_batch_2', 'd4bba439e000b95fd0a9bffe97cbabec'],
        ['data_batch_3', '54ebc095f3ab1f0389bbae665268c751'],
        ['data_batch_4', '634d18415352ddfa80567beed471001a'],
        ['data_batch_5', '482c414d41f54cd18b22e5b47cb7c3cb'],
    ]

    test_list = [
        ['test_batch', '40351d587109b95175f43aff81a1287e'],
    ]
    meta = {
        'filename': 'batches.meta',
        'key': 'label_names',
        'md5': '5ff9c542aee3614f3951f8cda6e48888',
    }

    def __init__(self,
                 train: bool,
                 download: bool,
                 data_folder: str,
                 transform: Optional[Union[BasicTransform, BaseCompose]],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 image_dtype: str = 'float32',
                 grayscale: bool = False,
                 test_mode: bool = False):
        """Init CIFAR10.

        Args:
            train: If True, train dataset will be used, else - test dataset.
            download: If True, data will be download and save to data_folder.
            data_folder: Directory with all the images.
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations` library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations` library.
            image_dtype: Data type of of the torch tensors related to the image.
            grayscale: If True, image will be read as grayscale otherwise as RGB.
            test_mode: If True, only image without labels will be returned.

        Raises:
            RuntimeError: if dataset or metadata file not found or corrupted.
        """
        super().__init__(transform, augment, image_dtype, grayscale, test_mode)
        self.__data_folder = Path(data_folder)
        self.__train = train

        if download:
            self.__download()

        if not self._check_integrity():
            raise RuntimeError('Dataset not found or corrupted. You can use download=True to download it')

        if self.__train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.__images = []
        self.__targets = []

        for file_name, _ in downloaded_list:
            file_path = self.__data_folder / self.base_folder / file_name
            with open(file_path, 'rb') as f:
                entry = pickle.load(f, encoding='latin1')
                self.__images.append(entry['data'])
                if 'labels' in entry:
                    self.__targets.extend(entry['labels'])
                else:
                    self.__targets.extend(entry['fine_labels'])

        self.__targets = np.array(self.__targets, dtype=np.int64)
        self.__images = np.vstack(self.__images).reshape(-1, 3, 32, 32)
        self.__images = self.__images.transpose((0, 2, 3, 1))  # convert to HWC

        self._load_meta()

    def _load_meta(self) -> None:
        """Load metadata."""
        path = self.__data_folder / self.base_folder / self.meta['filename']
        if not check_integrity(path, self.meta['md5']):
            raise RuntimeError('Dataset metadata file not found or corrupted. You can use download=True to download it')
        with open(path, 'rb') as infile:
            data = pickle.load(infile, encoding='latin1')
            self.classes = data[self.meta['key']]
        self.class_to_idx = {_class: i for i, _class in enumerate(self.classes)}

    def __getitem__(self, idx: int) -> dict:
        """Get item sample.

        Returns:
            sample: dict, where
            sample['image'] - Tensor, representing image after augmentations and transformations, dtype=image_dtype.
            sample['target'] - Target class or labels, dtype=target_dtype.
            sample['index'] - Index.
        """
        image = self.__images[idx]
        sample = {"image": image}
        sample = self._apply_transform(self.augment, sample)
        sample = self._apply_transform(self.transform, sample)
        sample['image'] = sample['image'].type(torch.__dict__[self._image_dtype])
        sample['index'] = idx

        if self._test_mode:
            return sample

        sample['target'] = self.__targets[idx]

        return sample

    def __len__(self) -> int:
        """Dataset length."""
        return len(self.__images)

    def _check_integrity(self) -> bool:
        """Check integrity."""
        for fentry in (self.train_list + self.test_list):
            filename, md5 = fentry[0], fentry[1]
            fpath = self.__data_folder / self.base_folder / filename
            if not check_integrity(fpath, md5):
                return False
        return True

    def __download(self) -> None:
        """Download archive by url to specific folder."""
        if self._check_integrity():
            print('Files already downloaded and verified')
        else:
            download_and_extract_archive(self.url, self.__data_folder, filename=self.filename, md5=self.tgz_md5)



from torch.optim import (
    Adadelta, Adagrad, Adam, Adamax, AdamW, ASGD, LBFGS,
    RMSprop, Rprop, SGD, SparseAdam
)

from torchok.constructor import OPTIMIZERS
from . import adafactor

# TODO: add other fresh optimizers from PyTorch
OPTIMIZERS.register_class(Adadelta)
OPTIMIZERS.register_class(Adagrad)
OPTIMIZERS.register_class(Adam)
OPTIMIZERS.register_class(Adamax)
OPTIMIZERS.register_class(AdamW)
OPTIMIZERS.register_class(ASGD)
OPTIMIZERS.register_class(LBFGS)
OPTIMIZERS.register_class(RMSprop)
OPTIMIZERS.register_class(Rprop)
OPTIMIZERS.register_class(SGD)
OPTIMIZERS.register_class(SparseAdam)



from typing import Dict, List, Optional
from dataclasses import dataclass, field
from enum import Enum


# Phase utils
class Phase(Enum):
    TRAIN = 'train'
    VALID = 'valid'
    TEST = 'test'
    PREDICT = 'predict'


# Callbacks parameters
@dataclass
class CallbacksParams:
    name: str
    params: Optional[Dict] = field(default_factory=dict)


# Optimization parameters
@dataclass
class OptmizerParams:
    name: str
    params: Optional[Dict] = field(default_factory=dict)


@dataclass
class SchedulerPLParams:
    # See https://pytorch-lightning.readthedocs.io/en/latest/common/lightning_module.html#configure-optimizers
    # for more information.
    interval: Optional[str] = 'epoch'
    frequency: Optional[int] = 1
    monitor: Optional[str] = 'val_loss'
    strict: Optional[bool] = True
    name: Optional[str] = None


@dataclass
class SchedulerParams:
    name: str
    params: Optional[Dict] = field(default_factory=dict)
    pl_params: Optional[SchedulerPLParams] = field(default_factory=lambda: SchedulerPLParams())


@dataclass
class OptimizationParams:
    optimizer: OptmizerParams
    scheduler: Optional[SchedulerParams] = None


# Data parameters
@dataclass
class AugmentationParams:
    name: str
    params: Dict = field(default_factory=dict)


@dataclass
class DatasetParams:
    name: str
    params: Dict
    transform: List[AugmentationParams]
    augment: Optional[List[AugmentationParams]] = field(default_factory=list)


@dataclass
class DataParams:
    dataset: DatasetParams
    dataloader: Dict


# Losses parameters
@dataclass
class LossParams:
    name: str
    mapping: Dict[str, str]
    params: Optional[Dict] = field(default_factory=dict)
    tag: Optional[str] = None
    weight: Optional[float] = None


@dataclass
class JointLossParams:
    losses: List[LossParams]
    normalize_weights: bool = True


# Metric parameters
@dataclass
class MetricParams:
    name: str
    mapping: Dict[str, str]
    params: Optional[Dict] = field(default_factory=dict)
    phases: Optional[List[Phase]] = field(default_factory=lambda: [Phase.TRAIN, Phase.VALID, Phase.TEST, Phase.PREDICT])
    prefix: Optional[str] = None


# Task parameters
@dataclass
class TaskParams:
    name: str
    compute_loss_on_valid: bool = True
    params: Optional[Dict] = field(default_factory=dict)
    base_checkpoint: Optional[str] = None
    overridden_checkpoints: Optional[Dict[str, str]] = None
    exclude_keys: Optional[List[str]] = None


# Trainer parameters
@dataclass
class TrainerParams:
    enable_checkpointing: bool = True
    default_root_dir: Optional[str] = None
    gradient_clip_val: Optional[float] = None
    gradient_clip_algorithm: Optional[str] = None
    num_nodes: int = 1
    num_processes: Optional[int] = None  # TODO: Remove in 2.0
    devices: Optional[List[int]] = None
    gpus: Optional[List[int]] = None
    auto_select_gpus: bool = False
    tpu_cores: Optional[List[int]] = None  # TODO: Remove in 2.0
    ipus: Optional[int] = None  # TODO: Remove in 2.0
    enable_progress_bar: bool = True
    overfit_batches: float = 0.0
    track_grad_norm: float = -1
    check_val_every_n_epoch: Optional[int] = 1
    fast_dev_run: bool = False
    accumulate_grad_batches: Optional[Dict[int, int]] = None
    max_epochs: Optional[int] = None
    min_epochs: Optional[int] = None
    max_steps: int = -1
    min_steps: Optional[int] = None
    max_time: Optional[Dict[str, int]] = None
    limit_train_batches: Optional[float] = None
    limit_val_batches: Optional[float] = None
    limit_test_batches: Optional[float] = None
    limit_predict_batches: Optional[float] = None
    val_check_interval: Optional[float] = None
    log_every_n_steps: int = 50
    accelerator: Optional[str] = None
    strategy: Optional[str] = None
    sync_batchnorm: bool = False
    precision: int = 32
    enable_model_summary: bool = True
    weights_save_path: Optional[str] = None  # TODO: Remove in 1.8
    num_sanity_val_steps: int = 2
    resume_from_checkpoint: Optional[str] = None
    profiler: Optional[str] = None
    benchmark: Optional[bool] = None
    reload_dataloaders_every_n_epochs: int = 0
    auto_lr_find: bool = False
    replace_sampler_ddp: bool = True
    detect_anomaly: bool = False
    auto_scale_batch_size: bool = False
    amp_backend: str = "native"
    amp_level: Optional[str] = None
    move_metrics_to_cpu: bool = False
    multiple_trainloader_mode: str = "max_size_cycle"


# Checkpoint parameters
@dataclass
class CheckpointParams:
    filename: Optional[str] = None
    monitor: str = 'valid/loss'
    verbose: bool = False
    save_last: bool = False
    save_top_k: Optional[int] = 1
    save_weights_only: bool = False
    mode: str = 'min'
    auto_insert_metric_name: bool = False
    export_to_onnx: bool = False
    onnx_params: Dict = field(default_factory=dict)


# Logger
@dataclass
class LoggerParams:
    name: str
    log_dir: str
    experiment_name: str = 'default'
    create_datetime_log_subdir: bool = True
    params: Optional[Dict] = field(default_factory=dict)


# Config parameters
@dataclass
class ConfigParams:
    # TODO add Logger params
    task: TaskParams
    data: Dict[Phase, List[DataParams]]
    trainer: TrainerParams
    optimization: Optional[List[OptimizationParams]] = None
    joint_loss: Optional[JointLossParams] = None
    checkpoint: Optional[CheckpointParams] = None
    logger: Optional[LoggerParams] = None
    metrics: Optional[List[MetricParams]] = field(default_factory=list)
    callbacks: Optional[List[CallbacksParams]] = field(default_factory=list)
    resume_path: Optional[str] = None



from typing import Optional

import torch.nn as nn
from torch import Tensor


class Mlp(nn.Module):
    """ MLP as used in Vision Transformer, MLP-Mixer and related networks."""
    def __init__(
            self,
            in_features: int,
            hidden_features: Optional[int] = None,
            out_features: Optional[int] = None,
            act_layer: nn.Module = nn.GELU):
        """Init Mlp.
        
        Args:
            in_features: Input features.
            hidden_features: Hidden features.
            out_features: Output features.
            act_layer: Activation layer.
        """
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)

    def forward(self, x: Tensor) -> Tensor:
        """Forward method."""
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)
        return x



from torchok.models.poolings.classification.pooling import (
    Pooling,
)
from torchok.models.poolings.classification.linear import (
    PoolingLinear,
)


__all__ = [
    'Pooling',
    'PoolingLinear'
]



from torch.nn import (
    L1Loss, NLLLoss, NLLLoss2d, PoissonNLLLoss, GaussianNLLLoss, KLDivLoss, MSELoss, BCELoss,
    BCEWithLogitsLoss, HingeEmbeddingLoss, MultiLabelMarginLoss, SmoothL1Loss, HuberLoss, SoftMarginLoss,
    CrossEntropyLoss, MultiLabelSoftMarginLoss, CosineEmbeddingLoss, MarginRankingLoss, MultiMarginLoss,
    TripletMarginLoss, TripletMarginWithDistanceLoss, CTCLoss
)

from torchok.constructor import LOSSES
from .segmentation import dice, lovasz
from .representation import pairwise
from .common import cross_entropy, focal

LOSSES.register_class(L1Loss)
LOSSES.register_class(NLLLoss)
LOSSES.register_class(NLLLoss2d)
LOSSES.register_class(PoissonNLLLoss)
LOSSES.register_class(GaussianNLLLoss)
LOSSES.register_class(KLDivLoss)
LOSSES.register_class(MSELoss)
LOSSES.register_class(BCELoss)
LOSSES.register_class(BCEWithLogitsLoss)
LOSSES.register_class(HingeEmbeddingLoss)
LOSSES.register_class(MultiLabelMarginLoss)
LOSSES.register_class(SmoothL1Loss)
LOSSES.register_class(HuberLoss)
LOSSES.register_class(SoftMarginLoss)
LOSSES.register_class(CrossEntropyLoss)
LOSSES.register_class(MultiLabelSoftMarginLoss)
LOSSES.register_class(CosineEmbeddingLoss)
LOSSES.register_class(MarginRankingLoss)
LOSSES.register_class(MultiMarginLoss)
LOSSES.register_class(TripletMarginLoss)
LOSSES.register_class(TripletMarginWithDistanceLoss)
LOSSES.register_class(CTCLoss)



from torchmetrics.aggregation import CatMetric, MaxMetric, MeanMetric, MinMetric, SumMetric  # noqa: E402
from torchmetrics.classification import (  # noqa: E402
    AUC,
    AUROC,
    ROC,
    Accuracy,
    AveragePrecision,
    BinnedAveragePrecision,
    BinnedPrecisionRecallCurve,
    BinnedRecallAtFixedPrecision,
    CalibrationError,
    CohenKappa,
    ConfusionMatrix,
    CoverageError,
    F1Score,
    FBetaScore,
    HammingDistance,
    HingeLoss,
    JaccardIndex,
    KLDivergence,
    LabelRankingAveragePrecision,
    LabelRankingLoss,
    MatthewsCorrCoef,
    Precision,
    PrecisionRecallCurve,
    Recall,
    Specificity,
    StatScores,
)
from torchmetrics.image import (  # noqa: E402
    ErrorRelativeGlobalDimensionlessSynthesis,
    MultiScaleStructuralSimilarityIndexMeasure,
    PeakSignalNoiseRatio,
    SpectralAngleMapper,
    SpectralDistortionIndex,
    StructuralSimilarityIndexMeasure,
    UniversalImageQualityIndex,
)
from torchmetrics.regression import (  # noqa: E402
    CosineSimilarity,
    ExplainedVariance,
    MeanAbsoluteError,
    MeanAbsolutePercentageError,
    MeanSquaredError,
    MeanSquaredLogError,
    PearsonCorrCoef,
    R2Score,
    SpearmanCorrCoef,
    SymmetricMeanAbsolutePercentageError,
    TweedieDevianceScore,
    WeightedMeanAbsolutePercentageError,
)

from torchok.metrics.classification import *
from torchok.metrics.segmentation import *
from torchok.metrics.representation import (
    PrecisionAtKMeter,
    RecallAtKMeter,
    MeanAveragePrecisionAtKMeter,
    NDCGAtKMeter,
)
from torchok.metrics.metrics_manager import *

from torchok.constructor import METRICS
from torchok.metrics.metrics_manager import MetricsManager, MetricWithUtils

METRICS.register_class(AUC)
METRICS.register_class(AUROC)
METRICS.register_class(ROC)
METRICS.register_class(Accuracy)
METRICS.register_class(AveragePrecision)
METRICS.register_class(BinnedAveragePrecision)
METRICS.register_class(BinnedPrecisionRecallCurve)
METRICS.register_class(BinnedRecallAtFixedPrecision)
METRICS.register_class(CalibrationError)
METRICS.register_class(CohenKappa)
METRICS.register_class(ConfusionMatrix)
METRICS.register_class(CoverageError)
METRICS.register_class(F1Score)
METRICS.register_class(FBetaScore)
METRICS.register_class(HammingDistance)
METRICS.register_class(HingeLoss)
METRICS.register_class(JaccardIndex)
METRICS.register_class(KLDivergence)
METRICS.register_class(LabelRankingAveragePrecision)
METRICS.register_class(LabelRankingLoss)
METRICS.register_class(MatthewsCorrCoef)
METRICS.register_class(Precision)
METRICS.register_class(PrecisionRecallCurve)
METRICS.register_class(Recall)
METRICS.register_class(Specificity)
METRICS.register_class(StatScores)

METRICS.register_class(ErrorRelativeGlobalDimensionlessSynthesis)
METRICS.register_class(MultiScaleStructuralSimilarityIndexMeasure)
METRICS.register_class(PeakSignalNoiseRatio)
METRICS.register_class(SpectralAngleMapper)
METRICS.register_class(SpectralDistortionIndex)
METRICS.register_class(StructuralSimilarityIndexMeasure)
METRICS.register_class(UniversalImageQualityIndex)

METRICS.register_class(CosineSimilarity)
METRICS.register_class(ExplainedVariance)
METRICS.register_class(MeanAbsoluteError)
METRICS.register_class(MeanAbsolutePercentageError)
METRICS.register_class(MeanSquaredError)
METRICS.register_class(MeanSquaredLogError)
METRICS.register_class(PearsonCorrCoef)
METRICS.register_class(R2Score)
METRICS.register_class(SpearmanCorrCoef)
METRICS.register_class(SymmetricMeanAbsolutePercentageError)
METRICS.register_class(TweedieDevianceScore)
METRICS.register_class(WeightedMeanAbsolutePercentageError)


__all__ = [
    'MetricsManager',
    'MetricWithUtils',
    'PrecisionAtKMeter',
    'RecallAtKMeter',
    'MeanAveragePrecisionAtKMeter',
    'NDCGAtKMeter',
]



"""TorchOK ConvBnAct module."""
from typing import Optional

import torch
import torch.nn as nn


class ConvBnAct(nn.Module):
    """Combination of convolution, batchnorm and activation layers."""

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 padding: int,
                 stride: int = 1,
                 bias: bool = False,
                 act_layer: Optional[nn.Module] = nn.ReLU):
        """Init ConvBnAct.

        Args:
            in_channels: Input channels.
            out_channels: Output channels.
            kernel_size: Kernel size.
            padding: Padding.
            stride: Stride.
                default: 1
            bias: Bias.
                default: False
            act_layer: Activation layer.
                default: relu.
        """
        super().__init__()
        self.conv = nn.Conv2d(in_channels,
                              out_channels,
                              kernel_size=kernel_size,
                              stride=stride,
                              padding=padding,
                              bias=bias)
        self.bn = nn.BatchNorm2d(out_channels)
        self.act = act_layer(inplace=True) if act_layer is not None else None

    def forward(self, x: torch.Tensor):
        """Forward method."""
        x = self.conv(x)
        x = self.bn(x)

        if self.act is not None:
            x = self.act(x)

        return x



from typing import List, Union, Tuple

import torch
from torch import Tensor
import torch.nn.functional as F

from torchok.constructor import NECKS
from torchok.models.necks.base_neck import BaseNeck


@NECKS.register_class
class HRNetSegmentationNeck(BaseNeck):
    """HRNet neck for segmentation task. """
    def __init__(self, in_channels: Union[List[int], Tuple[int, ...]]):
        """Init HRNetSegmentationNeck.

        Args:
            in_channels: Input channels.
        """
        out_channels = sum(in_channels)
        super().__init__(in_channels, out_channels)

    def forward(self, features: List[Tensor]) -> List[Tensor]:
        """Forward method."""
        input_image, x0, x1, x2, x3 = features

        x0_h, x0_w = x0.size(2), x0.size(3)
        x1 = F.interpolate(x1, size=(x0_h, x0_w), mode='bilinear', align_corners=True)
        x2 = F.interpolate(x2, size=(x0_h, x0_w), mode='bilinear', align_corners=True)
        x3 = F.interpolate(x3, size=(x0_h, x0_w), mode='bilinear', align_corners=True)

        feats = torch.cat([x0, x1, x2, x3], 1)
        return [input_image, feats]



""" TorchOK DiceLoss module.
Adapted from https://github.com/BloodAxe/pytorch-toolbelt/blob/develop/pytorch_toolbelt/losses/dice.py
Copyright (c) Eugene Khvedchenya
Licensed under The MIT License [see LICENSE for details]
"""
from typing import List, Union, Tuple
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from torchok.constructor import LOSSES


__all__ = ["DiceLoss"]

BINARY_MODE = "binary"
MULTICLASS_MODE = "multiclass"
MULTILABEL_MODE = "multilabel"


def soft_dice_score(y_pred: torch.Tensor, y_true: torch.Tensor, smooth: float = 0,
                    eps: float = 1e-7, dims: Union[int, List[int], Tuple[int, ...]] = None) -> torch.Tensor:
    """Compute Dice score.

    Shape:
        - Input: :math:`(N, NC, *)` where :math:`*` means any number
            of additional dimensions
        - Target: :math:`(N, NC, *)`, same shape as the input
        - Output: scalar.

    Args:
        y_pred: Prediction tensor.
        y_true: Ground truth tensor.
        smooth: Smooth value.
        eps: Small value to not division by zero.
        dims: Dims for which Dice score will calculate.

    Returns:
        dice_score: Dice score.

    Raises:
        ValueError: If shape y_pred and y_true don't match.
    """
    if y_pred.size() != y_true.size():
        raise ValueError(f"Shapes of y_pred and y_true don't match: {y_pred.size(), y_true.size()}")
    if dims is not None:
        intersection = torch.sum(y_pred * y_true, dim=dims)
        cardinality = torch.sum(y_pred + y_true, dim=dims)
    else:
        intersection = torch.sum(y_pred * y_true)
        cardinality = torch.sum(y_pred + y_true)
    dice_score = (2.0 * intersection + smooth) / (cardinality.clamp_min(eps) + smooth)
    return dice_score


def to_tensor(x: Union[torch.Tensor, list, tuple], dtype: torch.dtype = None) -> torch.Tensor:
    """Create torch tensor from torch.Tensor, list, tuple types, with specific dtype.

    Args:
        x: Input to convert.
        dtype: Output tensor dtype.

    Returns:
        Tensor with specific dtype.

    Raises:
        ValueError: If x not in torch.Tensor, list, tuple types.
    """
    if isinstance(x, torch.Tensor):
        if dtype is not None:
            x = x.type(dtype)
        return x
    if isinstance(x, (list, tuple)):
        x = np.ndarray(x)
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x

    raise ValueError("Unsupported input type" + str(type(x)))


@LOSSES.register_class
class DiceLoss(nn.Module):
    """Implementation of Dice loss for image segmentation task. It supports binary, multiclass and multilabel cases"""
    def __init__(self, mode: str, classes: List[int] = None, log_loss: bool = False,
                 from_logits: bool = True, smooth: float = 0, eps: float = 1e-7):
        """Init DiceLoss.

        Args:
            mode: Metric mode {'binary', 'multiclass', 'multilabel'}.
            classes: Optional list of classes that contribute in loss computation.
                By default, all channels are included.
            log_loss: If True, loss computed as `-log(jaccard)`; otherwise `1 - jaccard`.
            from_logits: If True assumes input is raw logits.
            smooth: Smooth value.
            eps: Small epsilon for numerical stability.

        Raises:
            ValueError: If mode not in set={'binary', 'multiclass', 'multilabel'}.
            ValueError: If classes parameter is not None in binary mode.
        """
        assert mode in {BINARY_MODE, MULTILABEL_MODE, MULTICLASS_MODE}
        if mode not in {BINARY_MODE, MULTILABEL_MODE, MULTICLASS_MODE}:
            raise ValueError(f'DiceLoss initialize. Mode {mode} does not supper. Please choose one of from'
                             f'{[BINARY_MODE, MULTILABEL_MODE, MULTICLASS_MODE]}.')
        super().__init__()
        self.mode = mode
        if classes is not None:
            if mode == BINARY_MODE:
                raise ValueError('DiceLoss initialize. Masking classes is not supported with mode=binary')
            classes = to_tensor(classes, dtype=torch.long)

        self.classes = classes
        self.from_logits = from_logits
        self.smooth = smooth
        self.eps = eps
        self.log_loss = log_loss

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward method for Dice loss.

        Args:
            input: NxCxHxW if mode is multiclass or multilabel and NxHxW if mode is binary
            target: NxHxW

        Returns:
            loss: Dice loss value - scalar.

        Raises:
            ValueError: If shape input tensor not match with target.
        """
        if self.mode in {BINARY_MODE, MULTILABEL_MODE} and input.shape != target.shape:
            raise ValueError(f"Shapes of input {input.shape} and target {target.shape} tensors don't match!")
        elif self.mode == MULTICLASS_MODE and input[:, 0].shape != target.shape:
            raise ValueError(f"Shapes of input {input.shape} and target {target.shape} tensors don't match!")

        bs = input.shape[0]
        num_classes = input.shape[1] if self.mode in {MULTICLASS_MODE, MULTILABEL_MODE} else 1

        if self.from_logits:
            # Apply activations to get [0..1] class probabilities
            # Using Log-Exp as this gives more numerically stable result and does not cause vanishing gradient on
            # extreme values 0 and 1
            if self.mode == MULTICLASS_MODE:
                input = input.log_softmax(dim=1).exp()
            else:
                input = F.logsigmoid(input).exp()

        dims = (0, 2)

        if self.mode == BINARY_MODE:
            target = target.view(bs, 1, -1)
            input = input.view(bs, 1, -1)

        if self.mode == MULTICLASS_MODE:
            target = target.view(bs, -1)
            input = input.view(bs, num_classes, -1)

            target = F.one_hot(target, num_classes)  # N,H*W -> N,H*W, C
            target = target.permute(0, 2, 1)  # H, C, H*W

        if self.mode == MULTILABEL_MODE:
            target = target.view(bs, num_classes, -1)
            input = input.view(bs, num_classes, -1)

        scores = soft_dice_score(input, target.type_as(input), self.smooth, self.eps, dims=dims)

        if self.log_loss:
            loss = -torch.log(scores.clamp_min(self.eps))
        else:
            loss = 1 - scores

        # Dice loss is undefined for non-empty classes
        # So we zero contribution of channel that does not have true pixels
        # NOTE: A better workaround would be to use loss term `mean(y_pred)`
        # for this case, however it will be a modified jaccard loss

        mask = target.sum(dims) > 0
        loss *= mask.to(loss.dtype)

        if self.classes is not None:
            loss = loss[self.classes]

        return loss.mean()



from torchok.data.datasets.base import ImageDataset
from torchok.data.datasets import classification
from torchok.data.datasets import detection
from torchok.data.datasets import examples
from torchok.data.datasets import representation
from torchok.data.datasets import segmentation


__all__ = [
    'ImageDataset',
]
